# pylint: disable=missing-docstring,line-too-long,import-error,broad-exception-raised
# pylint:disable=unnecessary-ellipsis

import unittest
from dataclasses import dataclass
from json import dumps
from typing import Dict, NamedTuple, Optional, Self, Type, TypedDict, cast
from unittest.mock import MagicMock, call, patch
from urllib.parse import parse_qs, urlparse

import numpy
import pandas

import soil
from soil.data_structure import DS_NAMESPACE, DataStructure
from soil.storage.base_storage import BaseStorage


class MockHttpResponse(NamedTuple):
    """Soil configuration class"""

    status_code: int
    text: str


# pylint: disable=unused-argument
def mock_http_post(
    url: str,
    *,
    headers: Optional[Dict[str, str]] = None,
    json: Optional[Dict[str, str]] = None,
    timeout: int,
) -> MockHttpResponse:
    assert json is not None
    response = {"_id": "mock_id", "type": json["type"]}
    return MockHttpResponse(status_code=201, text=dumps(response))


# pylint: disable=unused-argument, too-many-return-statements
def mock_http_get(
    url: str, *, headers: Optional[Dict[str, str]] = None, timeout: int
) -> MockHttpResponse:
    url_parts = urlparse(url)
    query_params = parse_qs(url_parts.query)
    if (
        url_parts.path == "/v2/states/"
        and query_params["name"][0] == "non_existing_alias_but_valid_id"
    ):
        return MockHttpResponse(status_code=200, text=dumps([]))
    if (
        url_parts.path == "/v2/states/"
        and query_params["name"][0] == "invalid_alias_invalidid"
    ):
        return MockHttpResponse(status_code=200, text=dumps([]))
    if url_parts.path == "/v2/states/" and query_params["name"][0] == "valid_alias":
        return MockHttpResponse(status_code=200, text=dumps([{"_id": "alias_id"}]))
    if url_parts.path == "/v2/states/alias_id/":
        result = {"state": {"result_id": "non_existing_alias_but_valid_id"}}
        return MockHttpResponse(status_code=200, text=dumps(result))
    if url_parts.path == "/v2/results/non_existing_alias_but_valid_id/":
        return MockHttpResponse(
            status_code=200,
            text=dumps({"_id": "mock_id", "type": "data_structures.some.test.type"}),
        )
    if url_parts.path == "/v2/results/invalid_alias_invalidid/":
        return MockHttpResponse(
            status_code=404,
            text=dumps(
                {"message": "Document with id invalid_alias_invalidid not found"}
            ),
        )
    raise Exception("mock http case not found")


@patch("soil.api.session.get", side_effect=mock_http_get)
@patch("soil.api.session.post", side_effect=mock_http_post)
class TestData(unittest.TestCase):
    def test_upload_data_object_list(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data([1, 2, 3, 4])
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + "list.List"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/results/",
            json={
                "type": "soil.data_structures.predefined.list.List",
                "data": "[1, 2, 3, 4]",
                "metadata": None,
            },
            timeout=30,
        )
        assert mock_get.call_count == 0

    def test_upload_data_object_with_return_type(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        @dataclass(kw_only=True)
        class MockClass:
            class Metadata(TypedDict):
                test: str

            metadata: "MockClass.Metadata"
            storage: BaseStorage

            def serialize(self) -> BaseStorage:
                """Serializes the DS."""
                ...

            @classmethod
            def deserialize(
                cls: Type[Self],
                storage: BaseStorage,
                metadata: "MockClass.Metadata",
            ) -> Self:
                """Deserialize DS method."""
                ...

        data = cast(DataStructure, soil.data([1, 2, 3, 4], return_type=MockClass))
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + "list.List"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/results/",
            json={
                "type": "soil.data_structures.predefined.list.List",
                "data": "[1, 2, 3, 4]",
                "metadata": None,
            },
            timeout=30,
        )
        assert mock_get.call_count == 0

    def test_upload_data_object_dict(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data({"test": "dictionary", "value": 11})
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + "dict.Dict"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/results/",
            json={
                "type": "soil.data_structures.predefined.dict.Dict",
                "data": '{"test": "dictionary", "value": 11}',
                "metadata": None,
            },
            timeout=30,
        )
        assert mock_get.call_count == 0

    def test_upload_data_object_numpy(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data(numpy.array([[1, 2], [3, 4]]))
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + "ndarray.Ndarray"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/results/",
            json={
                "type": "soil.data_structures.predefined.ndarray.Ndarray",
                "data": "[[1, 2], [3, 4]]",
                "metadata": None,
            },
            timeout=30,
        )
        assert mock_get.call_count == 0

    def test_upload_data_object_pandas(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data(pandas.DataFrame(data={"col1": [1, 2], "col2": [3, 4]}))
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == DS_NAMESPACE + "pd_data_frame.PdDataFrame"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        mock_post.assert_called_once_with(
            "http://test_host.test/v2/results/",
            json={
                "type": "soil.data_structures.predefined.pd_data_frame.PdDataFrame",
                "data": '{"col1":{"0":1,"1":2},"col2":{"0":3,"1":4}}',
                "metadata": None,
            },
            timeout=30,
        )
        assert mock_get.call_count == 0

    def test_get_data_valid_id(self, mock_post: MagicMock, mock_get: MagicMock) -> None:
        data = soil.data("non_existing_alias_but_valid_id")
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == "data_structures.some.test.type"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        assert mock_post.call_count == 0
        assert mock_get.call_count == 2
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=non_existing_alias_but_valid_id&with-roles",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/results/non_existing_alias_but_valid_id/",
                    timeout=30,
                ),
            ]
        )

    def test_get_data_valid_alias(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        data = soil.data("valid_alias")
        data = soil.data("non_existing_alias_but_valid_id")
        assert data.id == "mock_id"
        assert data.sym_id is None
        assert data.dstype == "data_structures.some.test.type"
        assert object.__getattribute__(data, "data") is None
        assert object.__getattribute__(data, "metadata") is None
        assert data.pipeline is None
        assert mock_post.call_count == 0
        assert mock_get.call_count == 5
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=valid_alias&with-roles",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/alias_id/",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/results/non_existing_alias_but_valid_id/",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/states/?name=non_existing_alias_but_valid_id&with-roles",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/results/non_existing_alias_but_valid_id/",
                    timeout=30,
                ),
            ]
        )

    def test_get_data_invalid_alias(
        self, mock_post: MagicMock, mock_get: MagicMock
    ) -> None:
        self.assertRaises(Exception, soil.data, "invalid_alias_invalidid")
        assert mock_post.call_count == 0
        assert mock_get.call_count == 2
        mock_get.assert_has_calls(
            [
                call(
                    "http://test_host.test/v2/states/?name=invalid_alias_invalidid&with-roles",
                    timeout=30,
                ),
                call(
                    "http://test_host.test/v2/results/invalid_alias_invalidid/",
                    timeout=30,
                ),
            ]
        )



# pylint: disable=missing-docstring,line-too-long, unused-argument
# type: ignore

import unittest
from unittest.mock import MagicMock

from soil import decorator

train = MagicMock(name="train")
my_model = MagicMock(name="my_model")


def trainable(module):
    def dec(**model_params):
        def train_fn(**train_params):
            def inner_module(*data):
                return train(
                    *data, module=module, model_params=model_params, **train_params
                )

            return inner_module

        return train_fn

    return dec


class TestDecorator(unittest.TestCase):
    def setUp(self):
        train.reset_mock()
        my_model.reset_mock()

    def test_no_depth(self):
        with self.assertRaises(ValueError):
            decorator()(trainable)
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls(self):
        my_trainable = decorator(depth=3)(trainable)
        assert my_trainable(__show_calls=True) == (trainable, [])
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_2(self):
        my_trainable = decorator(depth=3)(trainable)
        assert my_trainable(__show_calls=True) == (trainable, [])
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_3(self):
        my_trainable = decorator(depth=3)(trainable)(my_model)
        assert my_trainable(__show_calls=True) == (trainable, [((my_model,), {})])
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_4(self):
        my_trainable = decorator(depth=3)(trainable)(my_model)(test=1)
        assert my_trainable(__show_calls=True) == (
            trainable,
            [((my_model,), {}), ((), {"test": 1})],
        )
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_5(self):
        my_trainable = decorator(depth=3)(trainable)(my_model)(test=1)(["bla"])
        assert my_trainable(__show_calls=True) == (
            trainable,
            [((my_model,), {}), ((), {"test": 1}), ((["bla"],), {})],
        )
        train.assert_not_called()
        my_model.assert_not_called()

    def test_show_calls_6(self):
        my_trainable = decorator(depth=3)(trainable)(my_model)(test=1)(test2="3")(
            "other"
        )
        # pylint:disable=protected-access
        assert my_trainable._extract_mock_name() == "train()"
        train.assert_called_once_with(
            "other", module=my_model, test2="3", model_params={"test": 1}
        )
        my_model.assert_not_called()



import unittest
from dataclasses import dataclass
from datetime import UTC, datetime
from json import dumps, loads
from unittest.mock import Mock, patch
from urllib.parse import parse_qs, urlparse

from soil import job
from soil.job import _Job
from soil.types import Experiment


class TestJob(unittest.TestCase):
    @patch("soil.job.create_experiment")
    def test_job_create(self, mock_create_experiment: Mock):
        mock_create_experiment.return_value = Experiment(
            _id="my_experiment",
            status={"my_output": "DONE"},
            experiment_status="DONE",
            app_id="my_app",
            outputs={"my_datastructure": "my_output"},
            created_at=1640995200 * 1000,
        )
        plan_mocked = Mock()
        data_object_mocked = Mock(
            sym_id="my_datastructure", pipeline=Mock(plan=plan_mocked)
        )
        my_job = job(
            data_object=data_object_mocked,
            group="job_group1",
            description={"unit_id": "my_unit", "target": "my_target"},
        )
        mock_create_experiment.assert_called_once_with(
            plan_mocked,
            experiment_group="job_group1",
            description={"unit_id": "my_unit", "target": "my_target"},
        )

        self.assertEqual(
            my_job,
            _Job(  # pyright:ignore[reportAbstractUsage]
                experiment_id="my_experiment",
                result_id="my_output",
                _group="job_group1",
                _created_at=datetime.fromtimestamp(1640995200, tz=UTC).isoformat(),
            ),
        )

    @patch("soil.api.session.post")
    def test_job_create_with_description(
        self,
        mock_post: Mock,
    ):
        mock_post.return_value = MockHttpResponse(
            status_code=200,
            text=dumps(
                {
                    "experiment": {
                        "_id": "my_experiment",
                        "outputs": {"my_datastructure": "my_output"},
                        "created_at": 1640995200 * 1000,
                    }
                }
            ),
        )

        plan_mocked = Mock()
        data_object_mocked = Mock(
            sym_id="my_datastructure", pipeline=Mock(plan=plan_mocked)
        )
        my_job = job(
            data_object=data_object_mocked,
            group="job_group1",
            description={"unit_id": "my_unit", "target": "my_target"},
        )
        self.assertEqual(
            my_job,
            _Job(  # pyright:ignore[reportAbstractUsage]
                experiment_id="my_experiment",
                result_id="my_output",
                _group="job_group1",
                _created_at=datetime.fromtimestamp(1640995200, tz=UTC).isoformat(),
            ),
        )

        mock_post.assert_called_once_with(
            "http://test_host.test/v2/experiments/",
            json={
                "experiment": {
                    "name": "",
                    "description": {"unit_id": "my_unit", "target": "my_target"},
                    "plan": plan_mocked,
                    "experiment_group": "job_group1",
                }
            },
            timeout=30,
        )

    @patch("soil.api.session.post")
    def test_job_create_no_description(
        self,
        mock_post: Mock,
    ):
        mock_post.return_value = MockHttpResponse(
            status_code=200,
            text=dumps(
                {
                    "experiment": {
                        "_id": "my_experiment",
                        "outputs": {"my_datastructure": "my_output"},
                        "created_at": 1640995200 * 1000,
                    }
                }
            ),
        )

        plan_mocked = Mock()
        data_object_mocked = Mock(
            sym_id="my_datastructure", pipeline=Mock(plan=plan_mocked)
        )
        my_job = job(
            data_object=data_object_mocked,
            group="job_group1",
        )
        self.assertEqual(
            my_job,
            _Job(  # pyright:ignore[reportAbstractUsage]
                experiment_id="my_experiment",
                result_id="my_output",
                _group="job_group1",
                _created_at=datetime.fromtimestamp(1640995200, tz=UTC).isoformat(),
            ),
        )

        mock_post.assert_called_once_with(
            "http://test_host.test/v2/experiments/",
            json={
                "experiment": {
                    "name": "",
                    "description": "",
                    "plan": plan_mocked,
                    "experiment_group": "job_group1",
                }
            },
            timeout=30,
        )


@dataclass
class MockHttpResponse:
    """Soil configuration class"""

    status_code: int
    text: str

    def json(self) -> dict:
        return loads(self.text)


# pylint: disable=unused-argument
def mock_http_patch(
    url: str,
    *,
    headers: dict[str, str] | None = None,
    json: dict[str, str] | None = None,
    timeout: int,
) -> MockHttpResponse:
    assert url == "http://test_host.test/v2/states/mock_id/"
    assert json == {"name": "backtest", "state": {}}
    return MockHttpResponse(status_code=200, text=dumps({"hello": json}))


# pylint: disable=unused-argument
def mock_http_post_(
    url: str,
    *,
    headers: dict[str, str] | None = None,
    json: dict[str, str] | None = None,
    timeout: int,
) -> MockHttpResponse:
    url_parts = urlparse(url)
    assert url_parts.path == "/v2/states/"
    assert json == {"name": "backtest", "state": {}}
    return MockHttpResponse(status_code=200, text=dumps({}))


# pylint: disable=unused-argument
def mock_http_get(
    url: str, *, headers: dict[str, str] | None = None, timeout: int
) -> MockHttpResponse:
    url_parts = urlparse(url)
    query_params = parse_qs(url_parts.query)
    if url_parts.path == "/v2/states/" and query_params["name"][0] == "backtest":
        return MockHttpResponse(
            status_code=200,
            text=dumps([{"_id": "mock_id", "name": "backtest", "state": "mock_state"}]),
        )
    raise Exception("mock http case not found")



import unittest
from unittest.mock import patch

from soil.aio import alerts


class TestAioAlerts(unittest.IsolatedAsyncioTestCase):
    @patch("soil.aio.alerts.create_event")
    async def test_event(self, mock_create_event):
        mock_create_event.return_value = {"status": "success"}

        result = await alerts.event("test_key", "test_value")

        mock_create_event.assert_called_once_with("test_key", "test_value")
        self.assertEqual(result, {"status": "success"})

    @patch("soil.aio.alerts.create_alert")
    async def test_alert(self, mock_create_alert):
        mock_create_alert.return_value = {"alert_id": "123"}

        alert_config = {"type": "test_alert", "threshold": 100}
        result = await alerts.alert(alert_config)

        mock_create_alert.assert_called_once_with(alert_config)
        self.assertEqual(result, {"alert_id": "123"})


if __name__ == "__main__":
    unittest.main()



"""Module for Elasticsearch storage"""
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Optional

from soil.storage.base_storage import BaseStorage


@dataclass
class Elasticsearch(BaseStorage):
    """Class that implements Elasticsearch storage.

    Attributes
    ----------
        index: str: The index name to store the data to. It will be
            automatically prefixed with the app_id.
    """

    index: str

    def search(
        self, body: Dict[str, Any], auto_scroll: bool = True, **kwargs: Any
    ) -> Any:
        """Perform a search in elasticsearch in the idnex self.index.

        Attributes
        ----------
            body: The body of the search
            auto_scroll: If an auto_scroll has to be done.

        Returns: if auto_scroll is true it will return a tuple with a generator
            with the results as first element and metadata as second element.
            Otherwise it will return everything together.
        """

    # def insert(self, document: Dict[Any, Any], doc_id: Optional[str] = None) -> None:
    #     """Insert a document in the index."""

    def create_index(self, schema: Optional[Dict[str, Any]] = None) -> None:
        """Creates an index with the schema or an empty schema.
        If the index already exists it does nothing.
        """

    def delete_index(self) -> None:
        """Deletes the index"""

    def bulk(self, actions: Iterable[Any]) -> Any:
        """Performs bulk operations to the index only."""



"""Module for BaseStorage"""
from dataclasses import dataclass
from typing import Any, Type, TypeVar

StorageClass = TypeVar("StorageClass")


@dataclass
class BaseStorage:
    """Abstract class that implements serialize and deserialize methods for storage classes."""

    def serialize(self) -> dict[str, Any]:
        """Serializes the storage object. In general it shouldn't be used from a module."""

    @classmethod
    def deserialize(
        cls: Type[StorageClass], serialized_storage_object: dict[str, Any]
    ) -> StorageClass:
        """Takes a serialized storage object and returns an instance."""



"""Package that contains calls for the alerts system."""
from typing import Any, Dict

from soil import api


def event(key: str, value: Any) -> Any:
    """Creates an event."""
    return api.create_event(key, value)


def alert(alert_config: Dict) -> Any:
    """Creates an alert."""
    return api.create_alert(alert_config)



"""soil.decorator package."""
from typing import Any, Callable, Dict, List, Optional, Tuple, Union


def decorator(depth: Optional[int] = None) -> Callable:
    """Decorator of decorators. It allows to serialize a
    decorator that is not fully executed.

    If the decorated function is executed with __show_calls it will return the lists
    of args and kwargs that ran until that moment.

    When the function is fully executed it will run normally.

    Attributes
    ----------
        depth: The number of times the decorator will return a function.
    """
    if depth is None:
        raise ValueError("The depth of the decorator must be specified.")

    def wrapper(actual_decorator: Callable) -> Callable:
        def wrapper_logger_fn(
            args_list: List[Tuple[Any, ...]], kwargs_list: List[Dict[Any, Any]]
        ) -> Callable:
            def arg_logger_fn(
                *args: Any, __show_calls: bool = False, **kwargs: Any
            ) -> Union[
                Callable, Tuple[Callable, List[Tuple[Tuple[Any, ...], Dict[Any, Any]]]]
            ]:
                if __show_calls:
                    return (actual_decorator, list(zip(args_list, kwargs_list)))
                new_args_list = [*args_list, args]
                new_kwargs_list = [*kwargs_list, kwargs]
                if depth == len(args_list):
                    # if depth is 0 it will run the function normally.
                    res = actual_decorator
                    for args, kwargs in zip(new_args_list, new_kwargs_list):
                        res = res(*args, **kwargs)
                    return res
                return wrapper_logger_fn(new_args_list, new_kwargs_list)

            return arg_logger_fn

        return wrapper_logger_fn([], [])

    return wrapper



# pylint: disable=missing-docstring,line-too-long,unused-argument,no-member,too-many-locals
# type: ignore

import unittest
from unittest.mock import Mock, patch

import soil
from soil import errors, modulify
from soil.data_structure import DataStructure


class List:
    # pylint:disable=too-few-public-methods
    pass


@modulify(output_types=lambda *input_types, **args: [List])
def merge_list(data1, data2):
    pass


@modulify()
def merge_list_with_signature(data1, data2) -> tuple[DataStructure]:
    pass


@modulify()
def merge_list_with_signature_error(data1, data2) -> DataStructure:
    pass


@modulify
def merge_list_with_signature_error_2(data1, data2):
    pass


@modulify
def split_list(data) -> tuple[list, list]:
    pass


@modulify
def one_to_one(data, **kwargs) -> tuple[list]:
    pass


def soil_data_patch(name):
    return DataStructure(name)


@patch("soil.data", side_effect=soil_data_patch)
@patch(
    "soil.utils.uuid4",
    side_effect=list(range(1111, 1111 * 1000, 1111)),
)
@patch("soil.api.create_experiment")
@patch("soil.api.get_experiment")
@patch("soil.api.get_experiment_logs")
@patch("soil.api.get_result_data")
@patch("soil.pipeline.sleep", new=lambda x: None)
class TestPipelines(unittest.TestCase):
    # pylint:disable=too-many-function-args

    @patch("soil.api.get_result")
    def test_pipeline_simple(
        self,
        get_result_mock,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        data_ref_1 = soil.data("data_ref_1")
        data_ref_2 = soil.data("data_ref_2")
        (result,) = merge_list(data_ref_1, data_ref_2)
        get_experiment_mock.side_effect = [
            {"experiment_status": "EXECUTING"},
            {"experiment_status": "DONE"},
        ]
        experiment_logs_mock.return_value = []

        _data = result.metadata

        assert len(experiment_logs_mock.call_args_list) == 2
        assert len(get_result_data_mock.call_args_list) == 0
        assert len(get_result_mock.call_args_list) == 1
        assert len(get_experiment_mock.call_args_list) == 2
        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.merge_list-2222",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": ["data_ref_1", "data_ref_2"],
                    "outputs": ["unit/soil/test_pipelines/merge_list-1111-0"],
                    "args": {},
                }
            ]
        )

    @patch("soil.api.get_result")
    @patch("soil.pipeline.time")
    def test_pipeline_timeout(
        self,
        time_mock,
        get_result_mock,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock: Mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        time_mock.side_effect = [0, 50, 3601, 3801]
        data_ref_1 = soil.data("data_ref_1")
        data_ref_2 = soil.data("data_ref_2")
        (result,) = merge_list(data_ref_1, data_ref_2)
        get_experiment_mock.return_value = {"experiment_status": "EXECUTING"}
        experiment_logs_mock.return_value = []

        with self.assertRaises(errors.ExperimentTimeoutError):
            result.pipeline.run(timeout=3800)
        self.assertEqual(get_experiment_mock.call_count, 3)

    @patch("soil.api.get_result")
    def test_pipeline_simple_with_signature(
        self,
        get_result_mock,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        *_unused_mocks,
    ):
        data_ref_1 = soil.data("data_ref_1")
        data_ref_2 = soil.data("data_ref_2")
        (result,) = merge_list_with_signature(data_ref_1, data_ref_2)
        get_experiment_mock.side_effect = [
            {"experiment_status": "EXECUTING"},
            {"experiment_status": "DONE"},
        ]
        experiment_logs_mock.return_value = []

        _data = result.metadata

        assert len(experiment_logs_mock.call_args_list) == 2
        assert len(get_result_data_mock.call_args_list) == 0
        assert len(get_result_mock.call_args_list) == 1
        assert len(get_experiment_mock.call_args_list) == 2
        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.merge_list_with_signature-2222",
                    "module": "unit.soil.test_pipelines.merge_list_with_signature",
                    "inputs": ["data_ref_1", "data_ref_2"],
                    "outputs": [
                        "unit/soil/test_pipelines/merge_list_with_signature-1111-0"
                    ],
                    "args": {},
                }
            ]
        )

    @patch("soil.api.get_result")
    def test_pipeline_simple_with_signature_raise_error(
        self,
        *_unused_mocks,
    ):
        data_ref_1 = soil.data("data_ref_1")
        data_ref_2 = soil.data("data_ref_2")
        with self.assertRaises(ValueError):
            merge_list_with_signature_error(data_ref_1, data_ref_2)

        with self.assertRaises(ValueError):
            merge_list_with_signature_error_2(data_ref_1, data_ref_2)

    def test_pipeline_split(
        self,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        data_ref_1 = soil.data("data_ref_1")
        data_ref_2 = soil.data("data_ref_2")
        (result1, result2) = split_list(data_ref_1, data_ref_2)
        get_experiment_mock.side_effect = [{"experiment_status": "DONE"}]
        experiment_logs_mock.return_value = []

        _data = result1.data
        _data = result2.data

        assert len(experiment_logs_mock.call_args_list) == 1
        assert len(get_result_data_mock.call_args_list) == 2
        assert len(get_experiment_mock.call_args_list) == 1

        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.split_list-2222",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["data_ref_1", "data_ref_2"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-1111-0",
                        "unit/soil/test_pipelines/split_list-1111-1",
                    ],
                    "args": {},
                }
            ]
        )

    def test_pipeline_split_merge(
        self,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        data_ref_1 = soil.data("data_ref_1")
        (res_1, res_2) = split_list(data_ref_1)
        (result1,) = merge_list(res_1, res_2)
        get_experiment_mock.side_effect = [{"experiment_status": "DONE"}]
        experiment_logs_mock.return_value = []

        _data = result1.get_data()

        assert len(experiment_logs_mock.call_args_list) == 1
        assert len(get_result_data_mock.call_args_list) == 1
        assert len(get_experiment_mock.call_args_list) == 1

        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.split_list-2222",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["data_ref_1"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-1111-0",
                        "unit/soil/test_pipelines/split_list-1111-1",
                    ],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.merge_list-4444",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": [
                        "unit/soil/test_pipelines/split_list-1111-0",
                        "unit/soil/test_pipelines/split_list-1111-1",
                    ],
                    "outputs": ["unit/soil/test_pipelines/merge_list-3333-0"],
                    "args": {},
                },
            ]
        )

    def test_pipeline_complex_split_merge(
        self,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        data_ref_1 = soil.data("data_ref_1")
        (res_1, res_2) = split_list(data_ref_1)
        (res_11, res_12) = split_list(res_1)
        (res_21, res_22) = split_list(res_2)
        (result1,) = merge_list(res_11, res_21)
        (result2,) = merge_list(res_12, res_22)
        (result,) = merge_list(result1, result2)
        get_experiment_mock.side_effect = [{"experiment_status": "DONE"}]
        experiment_logs_mock.return_value = []

        _data = result.get_data()

        assert len(experiment_logs_mock.call_args_list) == 1
        assert len(get_result_data_mock.call_args_list) == 1
        assert len(get_experiment_mock.call_args_list) == 1

        create_experiment_mock.assert_called_once_with(
            [
                {
                    "id": "unit.soil.test_pipelines.merge_list-8888",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": [
                        "unit/soil/test_pipelines/split_list-3333-0",
                        "unit/soil/test_pipelines/split_list-5555-0",
                    ],
                    "outputs": ["unit/soil/test_pipelines/merge_list-7777-0"],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.split_list-4444",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["unit/soil/test_pipelines/split_list-1111-0"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-3333-0",
                        "unit/soil/test_pipelines/split_list-3333-1",
                    ],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.split_list-2222",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["data_ref_1"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-1111-0",
                        "unit/soil/test_pipelines/split_list-1111-1",
                    ],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.split_list-6666",
                    "module": "unit.soil.test_pipelines.split_list",
                    "inputs": ["unit/soil/test_pipelines/split_list-1111-1"],
                    "outputs": [
                        "unit/soil/test_pipelines/split_list-5555-0",
                        "unit/soil/test_pipelines/split_list-5555-1",
                    ],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.merge_list-11110",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": [
                        "unit/soil/test_pipelines/split_list-3333-1",
                        "unit/soil/test_pipelines/split_list-5555-1",
                    ],
                    "outputs": ["unit/soil/test_pipelines/merge_list-9999-0"],
                    "args": {},
                },
                {
                    "id": "unit.soil.test_pipelines.merge_list-13332",
                    "module": "unit.soil.test_pipelines.merge_list",
                    "inputs": [
                        "unit/soil/test_pipelines/merge_list-7777-0",
                        "unit/soil/test_pipelines/merge_list-9999-0",
                    ],
                    "outputs": ["unit/soil/test_pipelines/merge_list-12221-0"],
                    "args": {},
                },
            ]
        )
        # print(create_experiment_mock.call_args_list)

    @patch("soil.alias.set_alias")
    def test_pipeline_iterations(
        self,
        mock_set_alias,
        get_result_data_mock,
        experiment_logs_mock,
        get_experiment_mock,
        create_experiment_mock,
        random_int_patch,
        mock_data_patch,
    ):
        result = soil.data("data_ref_1")

        for i in range(4):
            experiment_logs_mock.return_value = []
            create_experiment_mock.return_value = {
                "_id": "aaaa",
                "outputs": {
                    "unit/soil/test_pipelines/one_to_one-1111-0": "okok",
                    "unit/soil/test_pipelines/one_to_one-3333-0": "blabla",
                    "unit/soil/test_pipelines/one_to_one-5555-0": "blabla",
                    "unit/soil/test_pipelines/one_to_one-7777-0": "blabla",
                },
            }
            get_experiment_mock.side_effect = [{"experiment_status": "DONE"}]
            (result2,) = one_to_one(result, index=i)
            self.assertEqual(len(result2.pipeline.plan), 1)
            soil.alias("result", result2)

        for i in range(4):
            experiment_logs_mock.return_value = []
            create_experiment_mock.return_value = {
                "_id": "aaaa",
                "outputs": {
                    "unit/soil/test_pipelines/one_to_one-9999-0": "blabla",
                    "unit/soil/test_pipelines/one_to_one-12221-0": "blabla",
                    "unit/soil/test_pipelines/one_to_one-14443-0": "blabla",
                    "unit/soil/test_pipelines/one_to_one-16665-0": "blabla",
                },
            }
            get_experiment_mock.side_effect = [{"experiment_status": "DONE"}]
            (result,) = one_to_one(result, index=i)
            # This is correct results in soil are immutable
            self.assertEqual(len(result.pipeline.plan), i + 1)
            soil.alias("result", result)
